// =============================================================================
//   OpenGL1.5 - Freepascal API-Template (gl15_fpc_template.zip)
// =============================================================================
//   Copyright  2004 by DGL - http://www.delphigl.com
// =============================================================================
//   Contents of this file are subject to the GNU Public License (GPL) which can
//   be obtained here : http://opensource.org/licenses/gpl-license.php
// =============================================================================
//   History :
//    Version 1.0 - Initial Release                            (Sascha Willems)
// =============================================================================
//   Based on Phobeus' OpenGL12-API Template for Delphi!
// =============================================================================

{$mode DELPHI}
{$apptype GUI}

program gl15_fpc_template;

uses
  Windows,
  Messages,
  dglOpenGL;

const
  WND_TITLE    = 'FreePascal OpenGL1.5-API Template - www.delphigl.com';
  FPS_TIMER    = 1;                       // Timer zur FPS berechnung
  FPS_INTERVAL = 1000;                    // Zeit zwischen FPS-Updates

var
  h_Wnd       : HWND;                     // Handle aufs Fenster
  h_DC        : HDC;                      // Device Context
  h_RC        : HGLRC;                    // OpenGL Rendering Context
  Keys        : Array[0..255] of Boolean; // Tasten-Status des Keyboards
  FPSCount    : Integer = 0;              // Frame-Zhler (fr FPS berechnen)
  ElapsedTime : Integer;                  // Programmlaufzeit
  Finished    : Boolean;
  Rotation    : Single;
  ClientRect  : TRect;                    // Speicher den Clientbereich des Fensters

//==============================================================================
// IntToStr
//==============================================================================
// Konvertiert Integers in Strings und vermeidet die Benutzung von SysUtils
//==============================================================================
function IntToStr(Num : Integer) : String;
begin
Str(Num, result);
end;

//==============================================================================
// glDraw
//==============================================================================
// Hier findet das Rendern der Szene statt
//==============================================================================
procedure glDraw;
begin
// Viewport an Fenster anpassen
GetClientRect(h_Wnd, ClientRect);
glViewport(0, 0, ClientRect.Right, ClientRect.Bottom);
glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);
// Projektionsmatrix setzen
glMatrixMode(GL_PROJECTION);
glLoadIdentity;
gluPerspective(60, ClientRect.Right/ClientRect.Bottom, 1, 512);
// Modelansichtsmatrix setzen
glMatrixMode(GL_MODELVIEW);
glLoadIdentity;

// Szeneninhalt rendern
glTranslatef(0, 0, -6);
glRotatef(Rotation, 0,1,0);

glBegin(GL_TRIANGLES);
 glColor3f(1,0,0); glVertex3f(-2, -2, 0);
 glColor3f(0,1,0); glVertex3f( 2, -2, 0);
 glColor3f(0,0,1); glVertex3f( 0,  2, 0);
glEnd;

// Rotieren
Rotation := Rotation + 0.1;
if Rotation > 360 then
 Rotation := 0;
end;


//==============================================================================
// glInit
//==============================================================================
// Initialisierung von OpenGL
//==============================================================================
procedure glInit();
begin
// Lschfarbe setzen
glClearColor(0, 0, 0.5, 0);
// Tiefentest anschalten
glEnable(GL_DEPTH_TEST);
// Art des Tiefentests (GL_LEQUAL = Fragmente mit niedrigerem oder gleichem Z-Wert durchlassen)
glDepthFunc(GL_LEQUAL);
end;

//==============================================================================
// ProcessKeys
//==============================================================================
// Tastatur Eingaben verarbeiten
//==============================================================================
procedure ProcessKeys;
begin
if (keys[VK_ESCAPE]) then
 finished := true;
end;

//==============================================================================
// WndProc
//==============================================================================
// Nachrichten verarbeiten
//==============================================================================
function WndProc(hWnd: HWND; Msg: UINT;  wParam: WPARAM;  lParam: LPARAM): LRESULT; stdcall;
begin
  case (Msg) of
    WM_CREATE:
      begin
        // Hier kann Zeugs rein das beim Programmstart ausgefhrt werden soll
      end;
    WM_CLOSE:
      begin
        PostQuitMessage(0);
        Result := 0
      end;
    WM_KEYDOWN:       // Taste als 'pressed' markieren
      begin
        keys[wParam] := True;
        Result := 0;
      end;
    WM_KEYUP:         // Taste als 'up' markieren
      begin
        keys[wParam] := False;
        Result := 0;
      end;
    WM_SIZE:          // Gre anpassen
      begin
        //glResizeWnd(LOWORD(lParam),HIWORD(lParam));
        Result := 0;
      end;
    WM_TIMER :                     // Hier werden alle benutzten Timer behandelt
      begin
        if wParam = FPS_TIMER then
        begin
          FPSCount :=Round(FPSCount * 1000/FPS_INTERVAL);   // FPS berechnen...
          SetWindowText(h_Wnd, PChar(WND_TITLE + '   [' + IntToStr(FPSCount) + ' FPS]')); //... und ausgeben!
          FPSCount := 0;
          Result := 0;
        end;
      end;
    else
      Result := DefWindowProc(hWnd, Msg, wParam, lParam);    //  Standard-Behandlung aller anderen Messages
  end;
end;


//==============================================================================
// glKillWnd
//==============================================================================
// Freigabe aller Fenster, Zeiger, Variablen
//==============================================================================
procedure glKillWnd(Fullscreen : Boolean);
var
 dmScreenSettings : DEVMODE;
begin
// Wenn Vollbild in Standardauflsung zurckkehren
if Fullscreen then
 begin
 dmScreenSettings := devmode(nil^);
 ChangeDisplaySettings(dmScreenSettings, 0);
 ShowCursor(True);
 end;
// Freigabe des Gerte und Rendering Contexts.
DeactivateRenderingContext;
DestroyRenderingContext(h_RC);
ReleaseDC(h_Wnd, h_DC);
// Schliet das Fenster
if ((h_Wnd <> 0) and (not DestroyWindow(h_Wnd))) then
 begin
 MessageBox(0, 'Unable to destroy window!', 'Error', MB_OK or MB_ICONERROR);
 h_Wnd := 0;
 end;
// Entfernt window class Registrierung
if (not UnRegisterClass('OpenGL', hInstance)) then
 begin
 MessageBox(0, 'Unable to unregister window class!', 'Error', MB_OK or MB_ICONERROR);
 hInstance := 0;
 end;
end;


//==============================================================================
// glCreateWnd
//==============================================================================
// Erstellt ein Fenster mit Passendem OpenGL Rendering Context
//==============================================================================
function glCreateWnd(Width, Height : Integer; Fullscreen : Boolean; PixelDepth : Integer) : Boolean;
var
  wndClass         : TWndClass;         // Fenster Klasse
  dwStyle          : DWORD;              // Fenster Stil
  dwExStyle        : DWORD;            // Erweiterter Fenster Stil
  dmScreenSettings : DEVMODE;   // Bildschirm Einstellungen (fullscreen, etc...)
  PixelFormat      : TGLuint;         // OpenGL Einstellungen (Pixelformat)
  h_Instance       : HINST;           // aktuelle Instanz
begin
// Instanz fr's Fenster holen
h_Instance := GetModuleHandle(nil);
// Daten in wndClass lschen
ZeroMemory(@wndClass, SizeOf(wndClass));
// Setup der Fenster Klasse
with wndClass do
 begin
 style         := CS_HREDRAW or    // Neuzeichenen wenn Fenster-Breite gendert
                  CS_VREDRAW or    // Neuzeichenen wenn Fenster-Hhe gendert
                  CS_OWNDC;        // Device Context exlusiv
 lpfnWndProc   := @WndProc;        // WndProc wird als Window Procedure gesetzt
 hInstance     := h_Instance;
 hCursor       := LoadCursor(0, IDC_ARROW);
 lpszClassName := 'OpenGL';
 end;

// Fenster Klasse registrieren
if (RegisterClass(wndClass) = 0) then
  begin
  MessageBox(0, 'Failed to register the window class!', 'Error', MB_OK or MB_ICONERROR);
  Result := False;
  Exit
  end;

// Vollbild wenn Parameter Fullscreen = true
if Fullscreen then
 begin
 ZeroMemory(@dmScreenSettings, SizeOf(dmScreenSettings));
 // Bildschirm Einstellungen werden festgelegt
 with dmScreenSettings do
  begin
  dmSize       := SizeOf(dmScreenSettings);
  dmPelsWidth  := Width;                    // Fenster Breite
  dmPelsHeight := Height;                   // Fenster Hhe
  dmBitsPerPel := PixelDepth;               // Farbtiefe (32bit etc)
  dmFields     := DM_PELSWIDTH or DM_PELSHEIGHT or DM_BITSPERPEL;
  end;
 // Bilschirm Modus auf Vollbild setzen
 if (ChangeDisplaySettings(dmScreenSettings, CDS_FULLSCREEN) = DISP_CHANGE_FAILED) then
  begin
  MessageBox(0, 'Unable to switch to fullscreen!', 'Error', MB_OK or MB_ICONERROR);
  Fullscreen := False;
  end;
 end;

// Wenn wir immer noch im Vollbildmodus sind....
if (Fullscreen) then
 begin
 dwStyle := WS_POPUP or                // Popup Fenster Stil
            WS_CLIPCHILDREN            // Kein Zeichnen in Child Fenstern
            or WS_CLIPSIBLINGS;        // Kein Zeichnen in Sibling Fenstern
 dwExStyle := WS_EX_APPWINDOW;         // Fenster im Vordergrund
 ShowCursor(False);                    // Mauszeiger verstecken
 end
// Fr Normale Fenster
else
 begin
 dwStyle := WS_OVERLAPPEDWINDOW or     // berschneidung zulassen
            WS_CLIPCHILDREN or         // Kein Zeichnen in Child Fenstern
            WS_CLIPSIBLINGS;           // Kein Zeichnen in Sibling Fenstern
 dwExStyle := WS_EX_APPWINDOW or       // Fenster im Fordergrund
              WS_EX_WINDOWEDGE;        // Erhobener Rand
 end;

// Das oben definierte Fenster wird erstellt
h_Wnd := CreateWindowEx(dwExStyle,      // Erweiterter Fenster Stil
                        'OpenGL',       // Name der Klasse
                        WND_TITLE,      // Fenster Titel (caption)
                        dwStyle,        // Fenster Stil
                        0, 0,           // Fenster Position
                        Width, Height,  // Gre des Fensters
                        0,              // Keine Paren-Windows
                        0,              // Kein Men
                        h_Instance,     // die Instanz
                        nil);           // Kein Parameter fr WM_CREATE
if h_Wnd = 0 then
 begin
 glKillWnd(Fullscreen);
 MessageBox(0, 'Unable to create window!', 'Error', MB_OK or MB_ICONERROR);
 Result := False;
 Exit;
 end;

// Den Gerte Kontext unseres Fensters besorgen
h_DC := GetDC(h_Wnd);
if (h_DC = 0) then
 begin
 glKillWnd(Fullscreen);
 MessageBox(0, 'Unable to get a device context!', 'Error', MB_OK or MB_ICONERROR);
 Result := False;
 Exit;
 end;

// Renderkontext erstellen
h_RC := CreateRenderingContext(h_DC,                // An den Gertekontext des Fensters anhngen
                               [opDoubleBuffered],  // Doppelte Pufferun
                               PixelDepth,          // BPP fr Farbpuffer
                               24,                  // BPP fr Tiefenpuffer
                               0,                   // Keinen Stencilpuffer
                               0,                   // Keinen Akkumulationspuffer
                               0,                   // Keine Auxpuffer
                               0);                  // Layer 0
// Renderkontext aktivieren
ActivateRenderingContext(h_DC, h_RC);
// Initialisierung des Timers zur FPS-Berechnung
SetTimer(h_Wnd, FPS_TIMER, FPS_INTERVAL, nil);
// Das Fenster wird in Vordergrund gebracht
ShowWindow(h_Wnd, SW_SHOW);
SetForegroundWindow(h_Wnd);
SetFocus(h_Wnd);

glInit;
Result := True;
end;


//==============================================================================
// WinMain
//==============================================================================
// Main message loop for the application
//==============================================================================
function WinMain(hInstance : HINST; hPrevInstance : HINST;
                 lpCmdLine : PChar; nCmdShow : Integer) : Integer; stdcall;
var
  msg : TMsg;
  appStart : DWord;
begin
Finished := False;
// OpenGL Funktionspointer initialisieren (hier knnen auch die Mesa-DLLs benutzt werden)
InitOpenGL('opengl32.dll', 'glu32.dll');
// Das Programm wird initialisiert (Fenster erstellen)
if not glCreateWnd(800, 600, FALSE, 32) then
 begin
 Result := 0;
 exit;
 end;
// Die Zeit zum Programmstart nehmen
appStart := GetTickCount;
// Main message loop:
while not finished do
 begin
 if (PeekMessage(@msg, 0, 0, 0, PM_REMOVE)) then // Nach Windows-Messages suchen
  begin
  if (msg.message = WM_QUIT) then     // Falls WM_QUIT dabei ist sind wir fertig
   finished := True
  else
   begin                               // Ansonsten wird die Message an das Programm weitergegeben
   TranslateMessage(@msg);
   DispatchMessage(@msg);
   end;
  end
 else
  begin
  Inc(FPSCount);                      // FPS Zhler erhhen
  ElapsedTime := GetTickCount() - appStart;     // Programmlaufzeit berechnen
  glDraw;                           // Szene zeichnen
  SwapBuffers(h_DC);                  // Szene ausgeben
  ProcessKeys;                      // Tastatureingaben verarbeiten
  end;
 end;
glKillWnd(FALSE);
Result := msg.wParam;
end;


begin
WinMain( hInstance, hPrevInst, CmdLine, CmdShow );
end.
